@@ -1,4 +1,4 @@
-import importlib_metadata
+import sys
 
 """
 Version has unique source in pyproject.toml.
@@ -6,6 +6,11 @@ importlib fetches version from distribution metadata files
 (in dist-info or egg-info dirs).
 From Python 3.8, importlib_metadata is in standard library as importlib.metadata.
 """
-__version__ = importlib_metadata.version("snakefmt")
+if sys.version_info >= (3, 8):
+    from importlib import metadata
+else:
+    import importlib_metadata as metadata
+
+__version__ = metadata.version("snakefmt")
 
 DEFAULT_LINE_LENGTH = 88
