@@ -12,7 +12,6 @@ from snakefmt.logging import Warnings
 from snakefmt.parser.parser import Parser, comment_start
 from snakefmt.parser.syntax import (
     COMMENT_SPACING,
-    TAB,
     InlineSingleParam,
     Parameter,
     ParameterSyntax,
@@ -20,14 +19,26 @@ from snakefmt.parser.syntax import (
     SingleParam,
     Syntax,
 )
-from snakefmt.types import TokenIterator
+from snakefmt.types import TAB, TokenIterator
 
-triple_quote_matcher = re.compile(
-    r"^\s*(\w?\"{3}.*?\"{3})|^\s*(\w?'{3}.*?'{3})", re.DOTALL | re.MULTILINE
+TAB_SIZE = len(TAB)
+# This regex matches any number of consecutive strings; each can span multiple lines.
+full_string_matcher = re.compile(
+    r"^\s*(\w?([\"']{3}.*?[\"']{3})|([\"']{1}.*?[\"']{1}))$", re.DOTALL | re.MULTILINE
 )
 contextual_matcher = re.compile(
     r"(.*)^(if|elif|else|with|for|while)([^:]*)(:.*)", re.S | re.M
 )
+after_if_keywords = ("elif", "else")
+
+
+def is_all_comments(string):
+    return all(
+        map(
+            comment_start,
+            [s for s in string.splitlines(keepends=True) if s.strip(" \t")],
+        )
+    )
 
 
 class Formatter(Parser):
@@ -63,9 +74,9 @@ class Formatter(Parser):
             return
 
         if not from_python:
-            formatted = self.run_black_format_str(self.buffer, self.target_indent)
-            if self.target_indent > 0:
-                formatted = self.align_strings(formatted, self.target_indent)
+            formatted = self.run_black_format_str(self.buffer, self.block_indent)
+            if self.keyword_indent > 0:
+                formatted = self.align_strings(formatted, self.keyword_indent)
         else:
             # Invalid python syntax, eg lone 'else:' between two rules, can occur.
             # Below constructs valid code statements and formats them.
@@ -73,7 +84,7 @@ class Formatter(Parser):
             if re_match is not None:
                 callback_keyword = re_match.group(2)
                 used_keyword = (
-                    "if" if callback_keyword in {"elif", "else"} else callback_keyword
+                    "if" if callback_keyword in after_if_keywords else callback_keyword
                 )
                 condition = re_match.group(3)
                 if condition != "":
@@ -83,7 +94,7 @@ class Formatter(Parser):
                 to_format = (
                     f"{re_match.group(1)}{test_substitute}" f"{re_match.group(4)}pass"
                 )
-                formatted = self.run_black_format_str(to_format, self.target_indent)
+                formatted = self.run_black_format_str(to_format, self.block_indent)
                 re_rematch = contextual_matcher.match(formatted)
                 if condition != "":
                     callback_keyword += re_rematch.group(3)
@@ -93,20 +104,19 @@ class Formatter(Parser):
                 formatted_lines = formatted.splitlines(keepends=True)
                 formatted = "".join(formatted_lines[:-1])  # Remove the 'pass' line
             else:
-                formatted = self.run_black_format_str(self.buffer, self.target_indent)
-            code_indent = self.syntax.code_indent
-            if code_indent is not None:
-                formatted = textwrap.indent(formatted, f"{TAB * code_indent}")
+                formatted = self.run_black_format_str(self.buffer, self.block_indent)
+
+            if self.syntax is not None:
+                formatted = textwrap.indent(formatted, f"{TAB * self.block_indent}")
 
         # Re-add newline removed by black for proper parsing of comments
         if self.buffer.endswith("\n\n"):
             if comment_start(self.buffer.rstrip().splitlines()[-1]):
                 formatted += "\n"
         # Only stick together separated single-parm keywords when separated by comments
-        buffer_is_all_comments = all(map(comment_start, self.buffer.splitlines()))
-        if not buffer_is_all_comments:
+        if not is_all_comments(self.buffer):
             self.last_recognised_keyword = ""
-        self.add_newlines(self.target_indent, formatted, final_flush, in_global_context)
+        self.add_newlines(self.block_indent, formatted, final_flush, in_global_context)
         self.buffer = ""
 
     def process_keyword_context(self, in_global_context: bool):
@@ -123,7 +133,7 @@ class Formatter(Parser):
         self, param_context: ParameterSyntax, in_global_context: bool
     ):
         self.add_newlines(
-            param_context.target_indent - 1,
+            param_context.keyword_indent - 1,
             in_global_context=in_global_context,
             context=param_context,
         )
@@ -131,10 +141,37 @@ class Formatter(Parser):
         self.last_recognised_keyword = param_context.keyword_name
 
     def run_black_format_str(
-        self, string: str, target_indent: int, extra_spacing: int = 0
+        self,
+        string: str,
+        target_indent: int,
+        extra_spacing: int = 0,
+        no_nesting: bool = False,
     ) -> str:
+        """
+        `no_nesting`: black uses one line skips between code blocks, instead of two,
+        for nested code (e.g. inside 'if'). We emulate this by putting nested code
+        inside an artificial 'if' statement. Setting `no_nesting` to True means the code
+        will always be formatted with two line skips.
+        """
+        if target_indent > 0 and comment_start(string):
+            lines = string.splitlines()
+            if len(lines) > 1:
+                lines[1] = textwrap.dedent(lines[1])
+                string = "\n".join(lines)
+
+        artificial_nest = (
+            self.from_python
+            and target_indent > 0
+            and not is_all_comments(string)
+            and len(string.strip().splitlines()) > 1
+            and not no_nesting
+        )
+
+        if artificial_nest:
+            string = f"if x:\n{textwrap.indent(string, TAB)}"
+
         # reduce black target line length according to how indented the code is
-        current_line_length = target_indent * len(TAB)
+        current_line_length = (target_indent or 0) * TAB_SIZE
         black_mode = copy(self.black_mode)
         black_mode.line_length = max(
             0, black_mode.line_length - current_line_length + extra_spacing
@@ -142,9 +179,34 @@ class Formatter(Parser):
         try:
             fmted = black.format_str(string, mode=black_mode)
         except black.InvalidInput as e:
-            raise InvalidPython(
-                f"Got error:\n```\n{str(e)}\n```\n" f"while formatting code with black."
-            ) from None
+            err_msg = ""
+            # Not clear whether all Black errors start with 'Cannot parse' - it seems to
+            # in the tests I ran
+            match = re.search(r"(Cannot parse: )(?P<line>\d+)(.*)", str(e))
+            try:
+                next_token = next(self.snakefile)
+                self.snakefile.denext(next_token)
+            except StopIteration:
+                next_token = None
+            if match and next_token is not None:
+                # this is the line number within the piece of code that was passed to
+                # black, not necessarily the line number within the Snakefile
+                line_num = int(match.group("line"))
+                context_line_num = next_token.start[0] - len(string.splitlines())
+                total_line_num = context_line_num + line_num - 1
+                err_msg = match.group(1) + str(total_line_num) + match.group(3)
+            else:
+                err_msg = str(e) + (
+                    "\n\n(Note reported line number may be incorrect, as"
+                    " snakefmt could not determine the true line number)"
+                )
+            err_msg = f"Black error:\n```\n{str(err_msg)}\n```\n"
+            raise InvalidPython(err_msg) from None
+
+        if artificial_nest:
+            lines = fmted.splitlines(keepends=True)[1:]
+            s = "".join(lines).lstrip("\n")
+            fmted = textwrap.dedent(s)
         return fmted
 
     def align_strings(self, string: str, target_indent: int) -> str:
@@ -154,19 +216,48 @@ class Formatter(Parser):
         pos = 0
         used_indent = TAB * target_indent
         indented = ""
-        for match in re.finditer(triple_quote_matcher, string):
+        for match in re.finditer(full_string_matcher, string):
             indented += textwrap.indent(string[pos : match.start(1)], used_indent)
+            lagging_spaces = len(indented) - len(indented.rstrip(" "))
+            lagging_indent = (
+                TAB * int(lagging_spaces / TAB_SIZE)
+                if lagging_spaces % TAB_SIZE == 0
+                else ""
+            )
             match_slice = string[match.start(1) : match.end(1)].replace("\t", TAB)
             all_lines = match_slice.splitlines(keepends=True)
             first = textwrap.indent(textwrap.dedent(all_lines[0]), used_indent)
+            is_multiline_string = re.match(
+                r"[bfru]?\"\"\"|'''", first.lstrip(), flags=re.IGNORECASE
+            )
             indented += first
+
             if len(all_lines) > 2:
-                middle = textwrap.indent(
-                    textwrap.dedent("".join(all_lines[1:-1])), used_indent
-                )
+                if is_multiline_string:
+                    middle = "".join(all_lines[1:-1])
+                else:
+                    mid = "".join(all_lines[1:-1])
+                    dedent_mid = textwrap.dedent(mid)
+                    lagging_indent_lvl = lagging_spaces // TAB_SIZE
+                    if lagging_indent_lvl == 0:
+                        required_indent_lvl = target_indent
+                    else:
+                        current_indent_lvl = (len(mid) - len(mid.lstrip())) // TAB_SIZE
+                        required_indent_lvl = current_indent_lvl + target_indent
+
+                    required_indent = TAB * required_indent_lvl
+                    middle = textwrap.indent(
+                        dedent_mid,
+                        required_indent,
+                    )
                 indented += middle
             if len(all_lines) > 1:
-                last = textwrap.indent(textwrap.dedent(all_lines[-1]), used_indent)
+                if is_multiline_string:
+                    last = all_lines[-1]
+                else:
+                    last = textwrap.indent(
+                        textwrap.dedent(all_lines[-1]), used_indent + lagging_indent
+                    )
                 indented += last
             pos = match.end()
         indented += textwrap.indent(string[pos:], used_indent)
@@ -192,12 +283,16 @@ class Formatter(Parser):
             raise InvalidParameterSyntax(f"{parameter.line_nb}{val}") from None
 
         if inline_formatting or param_list:
-            val = val.replace("\n", "")  # collapse strings on multiple lines
+            val = " ".join(
+                val.rstrip().split("\n")
+            )  # collapse strings on multiple lines
         extra_spacing = 0
         if param_list:
             val = f"f({val})"
             extra_spacing = 3
-        val = self.run_black_format_str(val, target_indent, extra_spacing)
+        val = self.run_black_format_str(
+            val, target_indent, extra_spacing, no_nesting=True
+        )
         if param_list:
             match_equal = re.match(r"f\((.*)\)", val, re.DOTALL)
             val = match_equal.group(1)
@@ -221,7 +316,7 @@ class Formatter(Parser):
         return result
 
     def format_params(self, parameters: ParameterSyntax) -> str:
-        target_indent = parameters.target_indent
+        target_indent = parameters.keyword_indent
         used_indent = TAB * (target_indent - 1)
 
         p_class = parameters.__class__
@@ -282,9 +377,10 @@ class Formatter(Parser):
                 and issubclass(context.__class__, SingleParam)
             )
             if not self.no_formatting_yet and not collate_same_singleparamkeyword:
-                if cur_indent == 0:
+                after_if_statement = self.buffer.startswith(after_if_keywords)
+                if max(cur_indent, 0) in (0, None) and not after_if_statement:
                     self.result += "\n\n"
-                elif in_global_context:
+                elif in_global_context or after_if_statement:
                     self.result += "\n"
         if in_global_context:  # Deal with comments
             if self.lagging_comments != "":
