@@ -47,6 +47,9 @@ rule_properties = dict(
     shadow=Context(None, SingleParam),
     group=Context(None, SingleParam),
     cache=Context(None, InlineSingleParam),
+    handover=Context(None, InlineSingleParam),
+    default_target=Context(None, InlineSingleParam),
+    retries=Context(None, InlineSingleParam),
 )
 
 
@@ -62,6 +65,7 @@ class SnakeRule(Vocabulary):
         notebook=Context(None, SingleParam),
         wrapper=Context(None, SingleParam),
         cwl=Context(None, SingleParam),
+        template_engine=Context(None, SingleParam),
         **rule_properties
     )
 
@@ -72,6 +76,7 @@ class SnakeModule(Vocabulary):
         config=Context(None, SingleParam),
         skip_validation=Context(None, SingleParam),
         meta_wrapper=Context(None, SingleParam),
+        prefix=Context(None, SingleParam),
         replace_prefix=Context(None, SingleParam),
     )
 
@@ -108,4 +113,5 @@ class SnakeGlobal(Vocabulary):
         scattergather=Context(None, ParamList),
         module=Context(SnakeModule, KeywordSyntax),
         use=Context(SnakeUseRule, KeywordSyntax),
+        resource_scopes=Context(None, KeywordSyntax),
     )
