@@ -4,11 +4,9 @@ Code in charge of parsing and validating Snakemake syntax
 import tokenize
 from abc import ABC, abstractmethod
 from re import match as re_match
-from typing import NamedTuple, Optional
 
 from snakefmt.exceptions import (
     ColonError,
-    DuplicateKeyWordError,
     EmptyContextError,
     InvalidParameter,
     InvalidParameterSyntax,
@@ -20,7 +18,6 @@ from snakefmt.exceptions import (
 )
 from snakefmt.types import (
     COMMENT_SPACING,
-    TAB,
     Token,
     TokenIterator,
     col_nb,
@@ -29,7 +26,6 @@ from snakefmt.types import (
 )
 
 # ___Token parsing___#
-QUOTES = {'"', "'"}
 BRACKETS_OPEN = {"(", "[", "{"}
 BRACKETS_CLOSE = {")", "]", "}"}
 
@@ -174,24 +170,14 @@ class Syntax(ABC):
     Classes derived from it raise syntax errors when snakemake syntax is not respected
     """
 
-    class Status(NamedTuple):
-        """Communicates the result of parsing a chunk of code"""
-
-        token: Token
-        indent: int
-        buffer: str
-        eof: bool
-        pythonable: bool
-
     def __init__(
-        self, keyword_name: str, target_indent: int, snakefile: TokenIterator = None
+        self, keyword_name: str, keyword_indent: int, snakefile: TokenIterator = None
     ):
         self.keyword_name = keyword_name
         self.keyword_line = keyword_name
-        assert target_indent >= 0
-        self.target_indent = target_indent
-        self.cur_indent = max(self.target_indent - 1, 0)
-        self.code_indent = None
+        assert keyword_indent >= 0
+        self.keyword_indent = keyword_indent
+        self.cur_indent = max(self.keyword_indent - 1, 0)
         self.comment = ""
         self.token = None
 
@@ -216,17 +202,16 @@ class KeywordSyntax(Syntax):
     def __init__(
         self,
         keyword_name: str,
-        target_indent: int,
+        keyword_indent: int,
         snakefile: TokenIterator = None,
         incident_syntax: "KeywordSyntax" = None,
         from_python: bool = False,
         accepts_py: bool = False,
     ):
         self.enter_context = True
-        super().__init__(keyword_name, target_indent, snakefile)
+        super().__init__(keyword_name, keyword_indent, snakefile)
         self.processed_keywords = set()
         self.accepts_python_code = accepts_py
-        self.queriable = True
         self.from_python = from_python
 
         if incident_syntax is not None:
@@ -247,17 +232,20 @@ class KeywordSyntax(Syntax):
         identifier = r"[a-zA-Z_]\S*"
         use_syntax_regexp = (
             r"use rule (?:(?:{id})|\*)"
-            r"(?: from {id})?(?: as {id})?( with[ ]?:)?$".format(id=identifier)
-        )
+            r"(?: from {id})?(?: exclude {id}(?:\s*,\s*{id})*)?"
+            r"(?: as {id})?( with[ ]?:)?$"
+        ).format(id=identifier)
         use_ebnf_syntax = (
             '"use" "rule" (identifier | "*") '
-            '"from" identifier ["as" identifier] ["with" ":"]'
+            '"from" identifier '
+            '["exclude" identifier {"," identifier}] '
+            '["as" identifier] ["with" ":"]'
         )
         while not is_newline(self.token):
             if self.token.type == tokenize.COMMENT:
                 break
             # Tokenizing splits up '<identifier>*' into two tokens
-            if self.token.string != "*":
+            if self.token.string not in ("*", ","):
                 self.keyword_line += " "
             self.keyword_line += self.token.string
             try:
@@ -289,13 +277,8 @@ class KeywordSyntax(Syntax):
             ColonError(self.line_nb, self.token.string, self.keyword_line)
         self.token = next(snakefile)
 
-    def add_processed_keyword(self, token: Token, keyword: str, check_dup: bool = True):
-        if check_dup and keyword in self.processed_keywords:
-            raise DuplicateKeyWordError(
-                f"L{token.start[0]}: '{keyword}' specified twice."
-            )
-        if keyword != "rule":  # Allows anonymous rules
-            self.processed_keywords.add(keyword)
+    def add_processed_keyword(self, token: Token, keyword: str):
+        self.processed_keywords.add(keyword)
 
     def check_empty(self):
         if len(self.processed_keywords) == 0:
@@ -303,57 +286,6 @@ class KeywordSyntax(Syntax):
                 f"{self.line_nb}{self.keyword_name} has no keywords attached to it."
             )
 
-    @property
-    def effective_indent(self) -> int:
-        return max(0, self.cur_indent - self.target_indent)
-
-    def get_next_queriable(self, snakefile: TokenIterator) -> Syntax.Status:
-        """Produces the next word that could be a snakemake keyword,
-        and additional information in a :Syntax.Status:
-        """
-        buffer = ""
-        newline = False
-        pythonable = False
-        prev_token: Optional[Token] = Token(tokenize.NAME)
-        while True:
-            token = next(snakefile)
-            if token.type == tokenize.INDENT:
-                self.cur_indent += 1
-                prev_token = None
-                continue
-            elif token.type == tokenize.DEDENT:
-                if self.cur_indent > 0:
-                    self.cur_indent -= 1
-                prev_token = None
-                continue
-            elif token.type == tokenize.ENDMARKER:
-                return self.Status(token, self.cur_indent, buffer, True, pythonable)
-            elif token.type == tokenize.COMMENT:
-                if token.start[1] == 0:
-                    return self.Status(token, 0, buffer, False, pythonable)
-            elif is_newline(token):
-                self.queriable, newline = True, True
-                buffer += "\n"
-                prev_token = None
-                continue
-
-            # Records relative tabbing, used for python code formatting
-            if newline and not token.type == tokenize.COMMENT:
-                buffer += TAB * self.effective_indent
-
-            if token.type == tokenize.NAME and self.queriable:
-                self.queriable = False
-                return self.Status(token, self.cur_indent, buffer, False, pythonable)
-
-            if add_token_space(prev_token, token):
-                buffer += " "
-            prev_token = token
-            if newline:
-                newline = False
-            if not pythonable and token.type != tokenize.COMMENT:
-                pythonable = True
-            buffer += token.string
-
 
 class ParameterSyntax(Syntax):
     """Parses snakemake keywords that do not accept other keywords, eg 'input'"""
@@ -361,11 +293,11 @@ class ParameterSyntax(Syntax):
     def __init__(
         self,
         keyword_name: str,
-        target_indent: int,
+        keyword_indent: int,
         incident_vocab: Vocabulary,
         snakefile: TokenIterator,
     ):
-        super().__init__(keyword_name, target_indent, snakefile)
+        super().__init__(keyword_name, keyword_indent, snakefile)
         self.positional_params, self.keyword_params = list(), list()
         self.eof = False
         self.incident_vocab = incident_vocab
@@ -419,7 +351,7 @@ class ParameterSyntax(Syntax):
                 if not cur_param.is_empty() and col_nb(self.token) < cur_param.col_nb:
                     exit = True
             else:
-                exit = self.cur_indent < self.target_indent
+                exit = self.cur_indent < self.keyword_indent
             if exit:
                 self.flush_param(cur_param, skip_empty=True)
         return exit
@@ -428,7 +360,7 @@ class ParameterSyntax(Syntax):
         token_type = self.token.type
         # Eager treatment of comments: tag them onto params
         if token_type == tokenize.COMMENT and not self.in_brackets:
-            cur_param.add_comment(self.token.string, self.target_indent)
+            cur_param.add_comment(self.token.string, self.keyword_indent)
             return cur_param
         if is_newline(self.token):  # Special treatment for inline comments
             if not cur_param.is_empty():
@@ -445,9 +377,8 @@ class ParameterSyntax(Syntax):
         if token_type == tokenize.INDENT:
             self.cur_indent += 1
         elif token_type == tokenize.DEDENT:
-            if self.cur_indent > 0:
-                self.cur_indent -= 1
-        elif is_equal_sign(self.token) and not self.in_brackets:
+            self.cur_indent = max(self.cur_indent - 1, 0)
+        elif is_equal_sign(self.token) and not self.in_brackets and not self.in_lambda:
             cur_param.to_key_val_mode(self.token)
         elif is_comma_sign(self.token) and not self.in_brackets and not self.in_lambda:
             cur_param.fully_processed = True
